package de.mannheim.ids.sruws;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.ws.rs.Consumes;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.StreamingOutput;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.events.XMLEvent;

import org.apache.commons.lang3.ArrayUtils;
import org.apache.http.HttpEntity;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.utils.URIBuilder;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.xml.sax.SAXException;

import eu.clarin.weblicht.wlfxb.io.WLDObjector;
import eu.clarin.weblicht.wlfxb.tc.api.MatchedCorpus;
import eu.clarin.weblicht.wlfxb.tc.api.MatchesLayer;
import eu.clarin.weblicht.wlfxb.tc.api.SentencesLayer;
import eu.clarin.weblicht.wlfxb.tc.api.Token;
import eu.clarin.weblicht.wlfxb.tc.api.TokensLayer;
import eu.clarin.weblicht.wlfxb.tc.xb.TextCorpusStored;
import eu.clarin.weblicht.wlfxb.xb.WLData;

/**
 * @author margaretha
 * */

@Path("/")
public class SRU_WS{
	
	private String endpointUri;
	private String endpointBase = "de.mannheim.ids.sruws.endpoint.";
	private Corpus defaultCorpus;
	
	private static final int DEFAULT_MAX_RECORDS = 10;
	private static final String HITS_NS = "http://clarin.eu/fcs/dataview/hits";
    private static final String HITS_PREFIX = "hits";    
    private static final String FCS_NS = "http://clarin.eu/fcs/1.0";
    private static final String FCS_PREFIX = "fcs";    
    //private static final String FCS_KWIC_NS = "http://clarin.eu/fcs/1.0/kwic";
    //private static final String FCS_KWIC_PREFIX = "kwic";
        
    private static String CORPUS_LANG = "de";
    private static String QUERY_LANG = "CQL";   
	private Map<String,Corpus> corpusList;
	private SAXParserFactory factory;
	private Logger logger;
	
	@Context
	ServletConfig config;	
	
	public SRU_WS() {
		logger = (Logger) LoggerFactory.getLogger(SRU_WS.class);
		factory = SAXParserFactory.newInstance();
		corpusList = new HashMap<String,Corpus>();
	}
	
	@GET
	@Produces(MediaType.TEXT_XML)
	public StreamingOutput doGet(@QueryParam("query") String query, 
			@QueryParam("endpoint") String endpoint, 
			@QueryParam("maximumRecords") int maximumRecords) throws IOException, ServletException{
		return process(endpoint, query,  maximumRecords);
	};
	
	@POST
	@Produces(MediaType.TEXT_XML)
	@Consumes("text/query-cql")
	public StreamingOutput doPost(@QueryParam("query") String query, 
			@QueryParam("endpoint") String endpoint, 
			@QueryParam("maximumRecords") int maximumRecords) throws IOException, ServletException{
		
		return process(endpoint, query,  maximumRecords);				
	}
	
	private StreamingOutput process(String endpoint, String query, int max) 
			throws IOException, ServletException {		
		
		if (endpoint == null || endpoint.isEmpty()){
			throw new WebApplicationException(
					Response.status(Response.Status.BAD_REQUEST)
							.entity("Endpoint parameter cannot be empty.")
							.type(MediaType.TEXT_PLAIN).build());
		}
		else if (config.getInitParameter(endpointBase+endpoint) == null){
			throw new WebApplicationException(
					Response.status(Response.Status.BAD_REQUEST)
							.entity("Endpoint URL for the endpoint parameter cannot be found.")
							.type(MediaType.TEXT_PLAIN).build());
		}
		else if (query == null || query.isEmpty()){
			throw new WebApplicationException(
					Response.status(Response.Status.BAD_REQUEST)
							.entity("Query cannot be empty.")
							.type(MediaType.TEXT_PLAIN).build());
		}
					
		if (max==0) max = DEFAULT_MAX_RECORDS;
		
		endpointUri = config.getInitParameter(endpointBase+endpoint);
		logger.info("Endpoint URL {} ",endpointUri);
		
		generateCorpusList();
		logger.info(query.toString());
		
		HttpGet getReq = createSearchRetrieveRequest(query, max);			
		CloseableHttpClient httpClient = HttpClients.createDefault();
		CloseableHttpResponse httpResponse = httpClient.execute(getReq);
		HttpEntity httpEntity = httpResponse.getEntity();
		BufferedInputStream bis = new BufferedInputStream(httpEntity.getContent());
		
		String text;
		try {
			text = readKwicInput(bis);
			//logger.info(text);
		} catch (XMLStreamException e) {				
			throw new ServletException("XML streaming error");
		}			
		
		return writeTCL(text, query);
		
	}
	
	private void generateCorpusList() throws IOException{
		corpusList = new HashMap<String, Corpus>();
		
		URI requestURI = null;
		try {
			URIBuilder builder = new URIBuilder(endpointUri);			
			builder.setParameter("operation", "scan");
			builder.setParameter("version", "1.2");
			builder.setParameter("scanClause", "fcs.resource=root");
			requestURI = builder.build();
		} catch (URISyntaxException e) {
			e.printStackTrace();
		}	
		HttpGet request = new HttpGet(requestURI);
		
		CloseableHttpClient httpClient = HttpClients.createDefault();
		CloseableHttpResponse httpResponse = httpClient.execute(request);
		HttpEntity httpEntity = httpResponse.getEntity();
		BufferedInputStream bis = new BufferedInputStream(httpEntity.getContent());			
		
		try {
			SAXParser saxParser = factory.newSAXParser();
			SRUScanResonseHandler handler = new SRUScanResonseHandler();
			saxParser.parse(bis, handler);
			corpusList = handler.getCorpora();					
			
		} catch (ParserConfigurationException | SAXException e) {
			e.printStackTrace();
		}
		
		defaultCorpus = new Corpus(endpointUri, endpointUri);
		corpusList.put(endpointUri, defaultCorpus);
	}
			
	private StreamingOutput writeTCL(String text, String query){
		final TextCorpusStored textCorpus = new TextCorpusStored(CORPUS_LANG);
		textCorpus.createTextLayer().addText(text);
		
		TokensLayer tokensLayer = textCorpus.createTokensLayer();
		SentencesLayer sentencesLayer = textCorpus.createSentencesLayer();	
		MatchesLayer matchesLayer = textCorpus.createMatchesLayer(QUERY_LANG, query);
		
		Corpus c;
		MatchedCorpus matchedCorpus; 
		for (String pid : corpusList.keySet()){
			c = corpusList.get(pid);			
			//logger.info(pid);
			List<Sentence> sentences = c.getSentences();
			if (sentences.size() == 0) continue;
			
			matchedCorpus = matchesLayer.addCorpus(c.getName(), c.getPid());			
			List<Token> itemToken = new ArrayList<Token>();			
			
			for (Sentence sentence : sentences){								
				List<Token> sentenceTokens = new ArrayList<Token>();
				
				String[] tokenStrings = sentence.getTokens();				
				int[] keyIndexes = sentence.getKeyIndexes();				
				
				int j=0; // counter for keyIndexes				
				for (int k = 0; k < tokenStrings.length; k++) {
					
					String tokenString = tokenStrings[k];
					Token token = tokensLayer.addToken(tokenString);
					sentenceTokens.add(token);	
										
					if (j < keyIndexes.length && keyIndexes[j] == k){
						itemToken.add(j,token);
						j++;
					}
				}		
				
				sentencesLayer.addSentence(sentenceTokens);		
				matchesLayer.addItem(matchedCorpus, itemToken);
				itemToken.clear();
			}
		}
		
		return new StreamingOutput() {			
			@Override
			public void write(OutputStream output) throws IOException,
					WebApplicationException {				
				WLData wlData = new WLData(textCorpus);
				WLDObjector.write(wlData, output);
				output.flush();
			}
		};
		
	}
	
	private String readKwicInput(InputStream is) throws XMLStreamException, UnsupportedEncodingException{
		
		String kwic;
		StringBuilder sb = null;
		StringBuilder kwicCollector = new StringBuilder();	
		
		XMLInputFactory f = XMLInputFactory.newInstance();
		XMLStreamReader reader = f.createXMLStreamReader(is);
		
		Corpus c = null;
		int e;
		while (reader.hasNext()) {
			e = reader.next();
			
			if (e == XMLEvent.START_ELEMENT &&
					reader.getLocalName().equals("Resource") &&
					reader.getPrefix().equals(FCS_PREFIX) &&
					reader.getNamespaceURI().equals(FCS_NS)){
				
				for (int i=0; i<reader.getAttributeCount(); i++){
					if (reader.getAttributeLocalName(i).equals("pid")){
						String pid = reader.getAttributeValue(i);						
						logger.info("Corpus: " + URLDecoder.decode(pid,"UTF-8"));						
						c = corpusList.get(URLDecoder.decode(pid,"UTF-8"));
						if (c == null) c = defaultCorpus;
					}					
				}
			}
							
			else if (e == XMLEvent.START_ELEMENT &&
					reader.getLocalName().equals("kwic") ){
				sb = new StringBuilder();
			}
			
			else if (e == XMLEvent.START_ELEMENT &&
					reader.getLocalName().equals("c") ){					
				e = reader.next();
				while (e == XMLEvent.CHARACTERS){
					//logger.info("context: "+reader.getText());
					sb.append(reader.getText());
					e = reader.next();
				}
				sb.append(" ");
			}				

			else if (e == XMLEvent.START_ELEMENT &&
					reader.getLocalName().equals("kw")){				
				e = reader.next();
				if (e == XMLEvent.CHARACTERS){
					String keyword = reader.getText();
					sb.append("#");
					sb.append(keyword);
					sb.append("# ");	
				}
			}
			
			else if (e == XMLEvent.END_ELEMENT && 
					reader.getLocalName().equals("kwic")){
				
				kwic = normalizeString(sb.toString());
				Sentence s = createSentence(kwic);
				c.addSentence(s);
				//logger.info(kwic);
				kwicCollector.append(s.getSentenceString());				
			}				
		}
		return kwicCollector.toString();
	}	
	
	private String readInput(InputStream is, List<String[]> tokenizedSentences) 
			throws XMLStreamException{
		
		String kwic;
		StringBuilder sb;
		StringBuilder kwicCollector = new StringBuilder();	
		
		XMLInputFactory f = XMLInputFactory.newInstance();
		XMLStreamReader reader = f.createXMLStreamReader(is);
		
		while (reader.hasNext()) {			
			if (reader.next() == XMLEvent.START_ELEMENT &&
					reader.getLocalName().equals("Result") &&
					reader.getPrefix().equals(HITS_PREFIX) &&
					reader.getNamespaceURI().equals(HITS_NS)){
					
				sb = new StringBuilder();
				
				reader.next();					
				sb.append(reader.getText());
				sb.append(" "); // left context
				
				reader.next();
				if (reader.getLocalName().equals("Hit") &&
						reader.getPrefix().equals(HITS_PREFIX) &&
						reader.getNamespaceURI().equals(HITS_NS)){
					
					reader.next();
					sb.append(reader.getText());
					sb.append(" ");
				}
				reader.next(); // end element Hit					
				reader.next();
				sb.append(reader.getText()); // right context
				
				kwic = normalizeString(sb.toString());
				tokenizedSentences.add(kwic.split("\\s+"));
				kwicCollector.append(kwic);
			}			
		}
		return kwicCollector.toString();
	}
	
	private Sentence createSentence(String kwic){
		String[] tokens = kwic.split("\\s+");				
		int open = ArrayUtils.indexOf(tokens,"#");
		tokens = ArrayUtils.remove(tokens, open);
		int close = ArrayUtils.lastIndexOf(tokens,"#");				
		tokens= ArrayUtils.remove(tokens, close);
		
		int[] keyIndexes = new int[close-open];
		int j=0;
		for (int i=open; i<close; i++){
			//logger.info(i+" "+tokens[i]);
			keyIndexes[j] = i;
			j++;
		}
		
		StringBuilder sb = new StringBuilder();
		for (String t: tokens){
			sb.append(t);
			sb.append(" ");
		}
		String sentenceString = sb.toString();
		
		return new Sentence(sentenceString, tokens, keyIndexes);		
	}
	
	private String normalizeString(String text){
		boolean isPrevPunctuation = false;
		
		StringBuilder sb = new StringBuilder();		
		for (int j = 0; j < text.length(); j++){
			char temp = text.charAt(j);					
			if (!Character.isDigit(temp) && !Character.isLetter(temp)){						
				if (temp == '\''){ sb.append(temp);	}
				else { 
					sb.append(" ");
					isPrevPunctuation = true;
				}							
			}
			sb.append(temp);
			
			if (isPrevPunctuation) {
				sb.append(" ");
				isPrevPunctuation = false;
			}
		}
		return sb.toString().replaceAll("\\s+", " ");			
	}	
	
	private HttpGet createSearchRetrieveRequest(String query, int maximumRecords) 
			throws ServletException{
		try {
			URIBuilder builder = new URIBuilder(endpointUri);	
			builder.setParameter("operation", "searchRetrieve");
			builder.setParameter("version", "1.2");
			builder.setParameter("query", query);
			builder.setParameter("startRecord", "1");
			builder.setParameter("maximumRecords", String.valueOf(maximumRecords));
			URI requestURI = builder.build();
			return new HttpGet(requestURI);
		}
		catch (URISyntaxException e) {
			throw new ServletException();
		}
	}	
}
